package com.simpleguava.base;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

public class Consumers {
	
	private Consumers(){}
	
	// composite consumer 
	// -------------------------------------------------------------------
	
	public static <T> Consumer<T> composite(Consumer<T> first,Consumer<T> second){
		Preconditions.checkNotNull(first);
		Preconditions.checkNotNull(second);
		
		return new CompositeConsumer<T>(Arrays.asList(first,second));
	}
	
	public static <T> Consumer<T> composite(Iterable<Consumer<T>> consumerIterable){
		return new CompositeConsumer<T>(consumerIterable);
	}
	
	@SafeVarargs
	public static <T> Consumer<T> composite(Consumer<T>... consumers) {
		Preconditions.checkNotNull(consumers);
		List<Consumer<T>> list = new ArrayList<Consumer<T>>(consumers.length);
		for(Consumer<T> e : consumers){
			list.add(e);
		}
		return new CompositeConsumer<>(list);
	}
	
	static class CompositeConsumer<T> implements Consumer<T>{
		
		public final Iterable<Consumer<T>> mIterable;
		
		public CompositeConsumer(Iterable<Consumer<T>> iter){
			mIterable = Preconditions.checkNotNull(iter);
		}

		@Override
		public void accept(T value) {
			for(Consumer<T> consumer : mIterable){
				consumer.accept(value);
			}
		}
	}
	
	// empty consumer 
	// -----------------------------------------------------------------
	
	static class EmptyConsumer<T> implements Consumer<T>,Serializable {
		private static final long serialVersionUID = 2929400619835197778L;

		@Override
		public void accept(T value) {
		}
		
		// Preserves singleton property
        private Object readResolve() {
            return EMPTY_CONSUMER;
        }
        
        public boolean equals(Object o) {
            return (o instanceof Consumer) && (Consumer<?>)o == EMPTY_CONSUMER;
        }

        public int hashCode() { return 1; }
	}
	
	@SuppressWarnings("rawtypes")
	public static final EmptyConsumer EMPTY_CONSUMER = new EmptyConsumer();
	
	@SuppressWarnings("unchecked")
	public static final <T> Consumer<T> empty(){
		Collections.emptyList();
		return (Consumer<T>)EMPTY_CONSUMER;
	}

}
